
/*
 * This is a crude interpreter for the 4DL programming language.
 * See http://www.cliff.biffle.org/esoterica/4dl.html for details.
 *
 * Usage:
 *     4dl <4DL source file> [ print ]
 *   where you can append the keyword 'print' behind if you want the interpreter
 *   to print out the code, as it sees it.
 *
 * Implementation: Tomas Arni Jonasson (tomas [at] tomasarni [dot] net)
 *
 * Date: 30. December 2009
 *
 * Source file definitions:
 *  - Top line describes size of the four dimensions by integers, seperated by
 *    an 'x', i.e. '4x3x3x2' for x, y, z and t dimensions, respectively,
 *    followed by a blank line.
 *  - The x-axis lies from left to right, within each block. Each grid element
 *    is separated by extra space.
 *  - The t-axis also lies from left to right, however block-by-block to the
 *    right, each block separated by 3 spaces
 *  - The y-axis lies from top to bottom, within each block
 *  - The z-axis (similarly to the t-axis w.r.t. the x-axis) lies from top to
 *    bottom, through the blocks, each block separated by empty line.
 *  - E.g.:
 *      # Outputs a smiley
 *      # Author: Tomas Arni Jonasson (tomas [at] tomasarni [dot] net)
 *      
 *      3x3x2x2
 *      
 *      B B Z   Y ~ %
 *        ) ~   . ~ .
 *      ~ ~ ~   X . y
 *      
 *      ~ ~ Y   z . x
 *      ~ ; p   ~   p
 *      ~ ~ T   ~ ~ y
 *    which should output a " ;) " to your terminal.
 *      
 *
 * Minor modifications to the syntax definition:
 *  - Defining a flat-file source files, I needed an comfortable 'null'/empty
 *    space character. I chose tilde ('~'). It cannot therefore be used in any
 *    source code
 *  - To enable new-lines in flat file source code, I also had to reserve a nice
 *    character. I chose the grave accent ('`') for it is not often used in
 *    standard english ascii (which this program is designed to serve)
 *
 *
 */

#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <stdio.h>
#include <string.h>
#include <stack>
#include <vector>

using namespace std;

int main(int argc, char *argv[])
{
    // Variables
    int temp;
    string line;
    int i, j, index, index_old;
    int x, y, z, t;
    int a, b;
    bool found;
    int dim [5];
    // Stack
    stack<char,vector<char> > myStack;
    // Instruction pointer
    int ip [4] = {0, 0, 0, 0};
    int ip_m [4] = {1, 0, 0, 0};
    
    // Make sure the number of args is correct
    if (argc < 2 || argc > 3)
    {
        cout << "Usage: 4dl <4DL source file> [ print ]";
        return -1;
    }
    
    // Open the file
    ifstream f (argv[1]);
    
    // Get the dimensions
    line = "";
    while (line.length() == 0)
    {
        getline (f, line);
        if (line[0] == '#') line = "";
    }
    index = 0;
    for (i = 0; i < 4; i++)
    {
        // Find the x
        found = false;
        index_old = index;
        index = line.find ('x', index+1);
        dim[i] = atoi(line.substr(index_old, index - index_old).c_str());
        index++;
    }
    
    // Now, read the file accordingly
    char source [dim[0]][dim[1]][dim[2]][dim[3]];
    for (z = 0; z < dim[2]; z++) // next paragraph
    {
        getline (f, line); // Read over empty line
        
        for (y = 0; y < dim[1]; y++) // next row
        {
            index = 0;
            getline (f, line);
            
            for (t = 0; t < dim[3]; t++) // whole line
            {
                for (x = 0; x < dim[0]; x++) // part line
                {
                    source[x][y][z][t] = 0;
                    if (line[index] != '~')
                        source[x][y][z][t] = line[index];
                    index += 2;
                }
                // skip the extra space if not at end
                if (index == line.length())
                    break;
                index += 2;
            }
        }
    }
    
    // Close the file
    f.close();
    
    // Print the program, if applicable...
    if (argc == 3 && strcmp("print", argv[2]) == 0)
    {
        for (z = 0; z < dim[2]; z++) // next paragraph
        {
            for (y = 0; y < dim[1]; y++) // next row
            {
                for (t = 0; t < dim[3]; t++) // whole line
                {
                    for (x = 0; x < dim[0]; x++) // part line
                    {
                        if (source[x][y][z][t] != 0)
                            cout << source[x][y][z][t];
                        else
                            cout << (char)32;
                        cout << "~";
                    }
                    cout << "  ";
                }
                cout << endl;
            }
            if (z < dim[2] - 1)
                cout << endl;
        }
    }
    
    char s;
    while(true)
    {
        // Set some vars for convenience
        x = ip[0]; y = ip[1]; z = ip[2]; t = ip[3];
        s = source[x][y][z][t];
        // Check if source is empty
        if (s == 0)
        {
            for (i = 0; i < 4; i++) ip[i] += ip_m[i];
            continue;
        }
        // Parse input
        switch (s)
        {
            case 'p': myStack.push(source[x-1][y][z][t]); break;
            case 'P': myStack.push(source[x+1][y][z][t]); break;
            case 'b': myStack.push(source[x][y-1][z][t]); break;
            case 'B': myStack.push(source[x][y+1][z][t]); break;
            case 'd': myStack.push(source[x][y][z-1][t]); break;
            case 'D': myStack.push(source[x][y][z+1][t]); break;
            case 'q': myStack.push(source[x][y][z][t-1]); break;
            case 'Q': myStack.push(source[x][y][z][t+1]); break;
            case '+':
                a = (int)myStack.top(); myStack.pop();
                b = (int)myStack.top(); myStack.pop();
                myStack.push((char)(a+b));
                break;
            case '-':
                a = (int)myStack.top(); myStack.pop();
                b = (int)myStack.top(); myStack.pop();
                myStack.push((char)(b-a));
                break;
            case ',':
                cout << "Enter a character: ";
                cin >> a;
                myStack.push((char)a);
                break;
            case '.':
                a = (int)myStack.top(); myStack.pop();
                if ((char)a == '`')
                    cout << endl;
                else
                    cout << (char)a;
                break;
            case 'X': ip_m = {1, 0, 0, 0}; break;
            case 'x': ip_m = {-1, 0, 0, 0}; break;
            case 'Y': ip_m = {0, 1, 0, 0}; break;
            case 'y': ip_m = {0, -1, 0, 0}; break;
            case 'Z': ip_m = {0, 0, 1, 0}; break;
            case 'z': ip_m = {0, 0, -1, 0}; break;
            case 'T': ip_m = {0, 0, 0, 1}; break;
            case 't': ip_m = {0, 0, 0, -1}; break;
            case '?':
                a = (int)myStack.top(); myStack.pop();
                if (a != 0) // skip a coord
                    for (i = 0; i < 4; i++) ip[i] += ip_m[i];
                break;
            case '#': for (i = 0; i < 4; i++) ip[i] += ip_m[i];; break;
            case '0': myStack.push(0); break;
            case '2': 
                a = (int)myStack.top();// myStack.pop();
                myStack.push((char)a);
                break;
            case '%': goto LABEL_EXIT; break;
            default:
                cout << "BAD INPUT: '" << s << "' (" << (int)s << ")" << endl;
                break;
        }
        // Move ip
        for (i = 0; i < 4; i++) ip[i] += ip_m[i];
    }
    
LABEL_EXIT:
    return 0;
}
